package be.iminds.ilabt.jfed.experimenter_gui.ui;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.AdvertisementCounts;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.MaintenanceInfo;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerGlimpse;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.util.TimeUtil;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.collections.ListChangeListener;
import javafx.collections.MapChangeListener;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.Tooltip;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.util.Pair;
import org.controlsfx.glyphfont.Glyph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.*;

/**
 * User: twalcari
 * Date: 9/16/2015
 * Time: 14:52
 */
public class HealthStatusPane extends StackPane {
    private static final Logger LOG = LoggerFactory.getLogger(HealthStatusPane.class);

    private static final Comparator<? super MaintenanceInfo> MAINTENANCE_MESSAGE_COMPARATOR =
            Comparator.nullsLast(Comparator.comparing(MaintenanceInfo::getStart))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(MaintenanceInfo::getEnd)))
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(MaintenanceInfo::getPlanned)))
                    .thenComparing(MaintenanceInfo::getMessage)
                    .thenComparing(Comparator.nullsLast(Comparator.comparing(MaintenanceInfo::getReported)));

    @FXML
    private Label noInfoLabel;
    @FXML
    private GridPane infoGridPane;

    @FXML
    private Label healthKeyLabel;
    @FXML
    private Label healthLabel;
    @FXML
    private Label reputationLabel;
    @FXML
    private Label freeLabel;
    @FXML
    private Label freeInfoLabel;
    @FXML
    private Label ipFreeLabel;
    @FXML
    private Label ipFreeInfoLabel;
    @FXML
    private Label maintenanceLabel;

    @FXML
    private Glyph healthGlyph;
    @FXML
    private Glyph reputationGlyph;
    @FXML
    private AvailabilityMeter availabiltyMeter;
    @FXML
    private AvailabilityMeter ipAvailabiltyMeter;
    @FXML
    private Glyph maintenanceGlyph;

    @FXML
    private Button hwFreeInfoButton;

    @FXML
    private Button maintenanceInfoButton;

    @FXML
    private Label updateHealthDateLabel;
    @FXML
    private Label updateCountsDateLabel;
    @FXML
    private Button refreshButton;
    @FXML
    private Glyph refreshGlyph;

    private boolean showHwFree = false;

    private JFedGuiConfig config;
    private Server server;

    private final ObjectProperty<JFedGuiConfig.ResourceType> type = new SimpleObjectProperty<>(null);
//    private final ObjectProperty<ServerGlimpse> serverGlimpse = new SimpleObjectProperty<>(null);

    public HealthStatusPane() {
        FXMLUtil.injectFXML(this);

        availabiltyMeter.useVmCalculationMethodProperty().bind(type.isEqualTo(JFedGuiConfig.ResourceType.VM));
    }

    public void setConfig(@Nonnull JFedGuiConfig config) {
        reputationGlyph.setVisible(config.isShowReputationEnabled());
        this.config = config;

        if (server != null)
            setServer(server);
    }

    private Runnable unregisterServer;

    public void setServer(Server server) {
        this.server = server;
        if (unregisterServer != null)
            unregisterServer.run();

        if (config == null)
            return;

        boolean noInfo = true;
        if (config.getServerNodeCounts(server).getValue() != null &&
                config.getServerNodeCounts(server).getValue().getCountEntriesSize() > 0)
            noInfo = false;
        if (noInfo && !config.getServerHealthInfo(server).isBound())
            noInfo = false;
        if (noInfo && !config.getServerMaintenance(server).isEmpty())
            noInfo = false;
        if (noInfo && !config.getServerReputation(server).isBound())
            noInfo = false;
        if (noInfo) {
            noInfoLabel.setVisible(true);
            infoGridPane.setVisible(false);
            return;
        }
        noInfoLabel.setVisible(false);
        infoGridPane.setVisible(true);

        ChangeListener<ServerGlimpse.HealthInfo> healthListener =
                (observable, oldValue, newValue) -> showHealthInfo(newValue);
        config.getServerHealthInfo(server).addListener(healthListener);
        showHealthInfo(config.getServerHealthInfo(server).get());

        ChangeListener<Integer> reputationListener =
                (observable, oldValue, newValue) -> showReputation(newValue);
        config.getServerReputation(server).addListener(reputationListener);
        showReputation(config.getServerReputation(server).get());

        ListChangeListener<? super MaintenanceInfo> maintenanceChangeListener =
                (ListChangeListener.Change<? extends MaintenanceInfo> change) ->
                        showMaintenance(change.getList());
        config.getServerMaintenance(server).addListener(maintenanceChangeListener);
        showMaintenance(config.getServerMaintenance(server));

        ChangeListener<AdvertisementCounts> serverNodeCountChangeListener =
                (observable, oldValue, newValue) -> showCounts(newValue);
        config.getServerNodeCounts(server).addListener(serverNodeCountChangeListener);
        showCounts(config.getServerNodeCounts(server).get());

        unregisterServer = () -> {
            config.getServerHealthInfo(server).removeListener(healthListener);
            config.getServerReputation(server).removeListener(reputationListener);
            config.getServerMaintenance(server).removeListener(maintenanceChangeListener);
            config.getServerNodeCounts(server).removeListener(serverNodeCountChangeListener);
        };
    }

    public void showHealthInfo(@Nullable ServerGlimpse.HealthInfo healthInfo) {
        if (healthInfo == null) {
            healthLabel.setText("No Health info");
            return;
        }

        Tooltip healthTooltip;
        if (healthInfo.getHealthExplanationFull() != null)
            healthTooltip = new Tooltip(healthInfo.getHealthExplanationFull());
        else
            healthTooltip = new Tooltip("Health explanation unknown");

        healthLabel.setText(healthInfo.getHealth() != null ? healthInfo.getHealth() + "%" : "-");
        setGlyphColor(healthGlyph, healthInfo.getHealth());

        healthLabel.setTooltip(healthTooltip);
        healthKeyLabel.setTooltip(healthTooltip);
        healthGlyph.setTooltip(healthTooltip);

        String healthTime;
        if (healthInfo.getTimestamp() != null) {
            healthTime = TimeUtil.calcAgo(healthInfo.getTimestamp());
        } else {
            healthTime = "at unknown time";
        }

        updateHealthDateLabel.setText("Health updated "+healthTime+".");
    }

    public void showReputation(@Nullable Integer reputation) {
        reputationLabel.setText(reputation != null ? reputation + "%" : "-");
        setGlyphColor(reputationGlyph, reputation);
    }

    public void showCounts(@Nullable AdvertisementCounts counts) {
        LOG.debug("HealthStatusPane count update: counts==null -> {}  type.get()={}", counts == null, type.get());
        if (counts != null) {
            if (type.get() != null) {
                Integer countAvail;
                Integer countTotal;
                switch (type.get()) {
                    case RAW_PC: {
                        countAvail = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.RAW_PC, ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE);
                        countTotal = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.RAW_PC, ServerGlimpse.AdvertisementInfo.CountScope.ALL);
                        break;
                    }
                    case VM: {
                        countAvail = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.VM, ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE);
                        countTotal = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.VM, ServerGlimpse.AdvertisementInfo.CountScope.ALL);
                        break;
                    }
                    default: {
                        countAvail = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.ANY, ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE);
                        countTotal = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.ANY, ServerGlimpse.AdvertisementInfo.CountScope.ALL);
                        break;
                    }
                }
                LOG.debug("HealthStatusPane count update: {}/{}", countAvail, countTotal);
                if (countAvail != null && countTotal != null && countTotal != 0) {
                    freeLabel.setText("Free " + type.get().getFriendlyName() + ":");

                    availabiltyMeter.setAvailable(countAvail);
                    availabiltyMeter.setTotal(countTotal);
                    availabiltyMeter.setUndetermined(false);

                    if (!showHwFree)
                        freeInfoLabel.setText(countAvail + " of " + countTotal);
                    else {
                        LOG.debug("health is showing hw type counts: "+counts.getHwCountKeys());
                        String freeText =  countAvail + " of " + countTotal + "\n";
                        for (String hwType : counts.getHwCountKeys()) {
                            Integer allCount = counts.getHwCount(hwType, AdvertisementCounts.CountScope.ALL);
                            Integer availCount = counts.getHwCount(hwType, AdvertisementCounts.CountScope.AVAILABLE);
                            freeText += hwType + ": " + availCount + " of " + allCount + "\n";
                        }
                        freeInfoLabel.setText(freeText);
                    }
                    hwFreeInfoButton.setVisible(!counts.getHwCountKeys().isEmpty());
                    hwFreeInfoButton.setManaged(!counts.getHwCountKeys().isEmpty());
                } else {
                    freeLabel.setText("Free resources:");
                    availabiltyMeter.setAvailable(0);
                    availabiltyMeter.setTotal(0);
                    availabiltyMeter.setUndetermined(true);
                    freeInfoLabel.setText("-");
                    hwFreeInfoButton.setVisible(false);
                    hwFreeInfoButton.setManaged(false);
                }
            }

            Integer countIpv4Avail = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.IPV4, ServerGlimpse.AdvertisementInfo.CountScope.AVAILABLE);
            Integer countIpv4Total = counts.getCount(ServerGlimpse.AdvertisementInfo.CountedObject.IPV4, ServerGlimpse.AdvertisementInfo.CountScope.ALL);

            if (countIpv4Total != null && countIpv4Total > 0) {
                ipAvailabiltyMeter.setAvailable(countIpv4Avail);
                ipAvailabiltyMeter.setTotal(countIpv4Total);
                ipAvailabiltyMeter.setUndetermined(false);
                ipFreeInfoLabel.setText(countIpv4Avail + " of " + countIpv4Total);
            } else {
                ipAvailabiltyMeter.setAvailable(0);
                ipAvailabiltyMeter.setTotal(0);
                ipAvailabiltyMeter.setUndetermined(true);
                ipFreeInfoLabel.setText("-");
            }

            String countsTime;
            if (counts.getCountDate() != null) {
                countsTime = TimeUtil.calcAgo(counts.getCountDate());
            } else {
                countsTime = "at unknown time";
            }

            updateCountsDateLabel.setText("Counts updated "+countsTime+".");
        } else {
            ipAvailabiltyMeter.setAvailable(0);
            ipAvailabiltyMeter.setTotal(0);
            ipAvailabiltyMeter.setUndetermined(true);
            ipFreeInfoLabel.setText("-");

            freeLabel.setText("Free resources:");
            availabiltyMeter.setAvailable(0);
            availabiltyMeter.setTotal(0);
            availabiltyMeter.setUndetermined(true);
            freeInfoLabel.setText("-");
            hwFreeInfoButton.setVisible(false);
            hwFreeInfoButton.setManaged(false);

            updateCountsDateLabel.setText("Count info not available.");
        }
    }

    public void showMaintenance(@Nullable Collection<? extends MaintenanceInfo> maintenance) {
//        if (maintenance == null) {
////        //no maintenance:
//            maintenanceInfoButton.setVisible(false);
//            maintenanceLabel.setVisible(false);
//            maintenanceGlyph.setVisible(false);
//            return;
//        }
//        maintenanceInfoButton.setVisible(true);
//        maintenanceLabel.setVisible(true);
//        maintenanceGlyph.setVisible(true);

        Optional<? extends MaintenanceInfo> activeMaintenance;
        Optional<? extends MaintenanceInfo> upcomingMaintenance;
        if (maintenance != null && !maintenance.isEmpty()) {
            activeMaintenance =
                    maintenance.stream()
                            .filter(mm -> mm.getStartTime() == null || mm.getStartTime().isBefore(ZonedDateTime.now()))
                            .filter(mm -> mm.getEndTime() == null || mm.getEndTime().isAfter(ZonedDateTime.now()))
                            .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                            .findFirst();

            upcomingMaintenance =
                    maintenance.stream()
                            .filter(mm -> mm.getStartTime() != null && mm.getStartTime().isAfter(ZonedDateTime.now()))
                            .sorted(MAINTENANCE_MESSAGE_COMPARATOR)
                            .findFirst();
        } else {
            activeMaintenance = Optional.empty();
            upcomingMaintenance = Optional.empty();
        }
        if (activeMaintenance.isPresent()) {
            StringBuilder maintenanceMsg = new StringBuilder("Currently in maintenance ");

            ZonedDateTime startTime = activeMaintenance.get().getStartTime();
            if (startTime != null)
                maintenanceMsg.append("from ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(
                                startTime.withZoneSameInstant(ZoneId.systemDefault()))
                        )
                        .append(" ");
            ZonedDateTime endTime = activeMaintenance.get().getEndTime();
            if (endTime != null)
                maintenanceMsg.append("until ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(
                                endTime.withZoneSameInstant(ZoneId.systemDefault()))
                        );

            maintenanceLabel.setText(maintenanceMsg.toString());
            maintenanceGlyph.setTextFill(Color.RED);
        } else if (upcomingMaintenance.isPresent()) {
            StringBuilder maintenanceMsg = new StringBuilder("Upcoming maintenance ");

            ZonedDateTime startTime = upcomingMaintenance.get().getStartTime();
            if (startTime != null)
                maintenanceMsg.append("from ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(
                                startTime.withZoneSameInstant(ZoneId.systemDefault()))
                        )
                        .append(" ");
            ZonedDateTime endTime = upcomingMaintenance.get().getEndTime();
            if (endTime != null)
                maintenanceMsg.append("until ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(
                                endTime.withZoneSameInstant(ZoneId.systemDefault()))
                        );

            maintenanceLabel.setText(maintenanceMsg.toString());
            maintenanceGlyph.setTextFill(Color.ORANGE);
        } else {
            maintenanceLabel.setText("None planned");
            maintenanceGlyph.setTextFill(Color.GRAY);
        }

        maintenanceInfoButton.setDisable(!activeMaintenance.isPresent() && !upcomingMaintenance.isPresent());
    }

    @FXML
    private void refresh() {
        //refresh is not that usefull, as the time shown is that ONE THE SERVER.
        //   therefor, a user will be confused after a refresh...
        if (config != null) {
            LOG.debug("Updating ServerGlimpses");
            config.forceRefreshServerGlimpses();
//            updateDateLabel.setText(updateDateLabel.getText()+"...");
        } else {
            LOG.warn("Internal error: Cannot refresh");
//            updateDateLabel.setText("(Internal error: Cannot refresh)");
        }
    }

    @FXML
    private void toggleHwFree() {
        showHwFree = !showHwFree;

        showCounts(config.getServerNodeCounts(server).get());
    }

    @FXML
    private void showMaintenanceInfo() {
        ObservableList<MaintenanceInfo> maintenances = config.getServerMaintenance(server);
        assert maintenances != null;
        assert !maintenances.isEmpty();

        List<MaintenanceInfo> sortedMaintenances = new ArrayList<>(maintenances);
        sortedMaintenances.sort(MAINTENANCE_MESSAGE_COMPARATOR);


        Alert infoAlert = new Alert(Alert.AlertType.INFORMATION);
        infoAlert.setHeaderText("Planned maintenance");
        infoAlert.setContentText(formatMaintenanceInfos(sortedMaintenances));
        infoAlert.setResizable(true);
        infoAlert.showAndWait();
    }

    private static void setGlyphColor(@Nonnull Glyph glyph, @Nullable Integer value) {
        if (value == null) {
            glyph.setTextFill(Color.GRAY);
        } else if (value > 95) {
            glyph.setTextFill(Color.GREEN);
        } else if (value > 66) {
            glyph.setTextFill(Color.YELLOWGREEN);
        } else if (value > 33) {
            glyph.setTextFill(Color.ORANGE);
        } else {
            glyph.setTextFill(Color.RED);
        }
    }

    public JFedGuiConfig.ResourceType getType() {
        return type.get();
    }

    public ObjectProperty<JFedGuiConfig.ResourceType> typeProperty() {
        return type;
    }

    public void setType(JFedGuiConfig.ResourceType type) {
        this.type.set(type);
    }



    @Nonnull
    public static String formatMaintenanceInfos(List<MaintenanceInfo> maintenanceInfos) {
        StringBuilder maintenanceMsg = new StringBuilder();

        for (MaintenanceInfo activeMaintenance : maintenanceInfos) {
            if (activeMaintenance.getStartTime() != null && activeMaintenance.getEndTime() != null) {
                maintenanceMsg.append("From ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(activeMaintenance.getStartTime().withZoneSameInstant(ZoneId.systemDefault())))
                        .append(" until ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(activeMaintenance.getEndTime().withZoneSameInstant(ZoneId.systemDefault())))
                        .append(": ");
            } else if (activeMaintenance.getStartTime() != null) {
                maintenanceMsg.append("From ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(activeMaintenance.getStartTime().withZoneSameInstant(ZoneId.systemDefault())))
                        .append(": ");
            } else if (activeMaintenance.getEndTime() != null) {
                maintenanceMsg
                        .append("Until ")
                        .append(DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(activeMaintenance.getEndTime().withZoneSameInstant(ZoneId.systemDefault())))
                        .append(": ");
            }

            maintenanceMsg.append(activeMaintenance.getMessage()).append("\n\n");
        }
        return maintenanceMsg.toString().trim();
    }
}
